//          kname        modmin       modadd       ch  key
MLE_KEY_DEF("space",     TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_TILDE)
MLE_KEY_DEF("~",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_TILDE)
MLE_KEY_DEF("2",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_TILDE)
MLE_KEY_DEF("a",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_A)
MLE_KEY_DEF("b",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_B)
MLE_KEY_DEF("c",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_C)
MLE_KEY_DEF("d",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_D)
MLE_KEY_DEF("e",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_E)
MLE_KEY_DEF("f",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_F)
MLE_KEY_DEF("g",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_G)
MLE_KEY_DEF("h",         TB_MOD_CTRL, 0,           0,  TB_KEY_BACKSPACE)
MLE_KEY_DEF("i",         TB_MOD_CTRL, 0,           0,  TB_KEY_TAB)
MLE_KEY_DEF("tab",       0,           TB_MOD_CTRL, 0,  TB_KEY_TAB)
MLE_KEY_DEF("j",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_J)
MLE_KEY_DEF("k",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_K)
MLE_KEY_DEF("l",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_L)
MLE_KEY_DEF("m",         TB_MOD_CTRL, 0,           0,  TB_KEY_ENTER)
MLE_KEY_DEF("enter",     0,           TB_MOD_CTRL, 0,  TB_KEY_ENTER)
MLE_KEY_DEF("n",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_N)
MLE_KEY_DEF("o",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_O)
MLE_KEY_DEF("p",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_P)
MLE_KEY_DEF("q",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_Q)
MLE_KEY_DEF("r",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_R)
MLE_KEY_DEF("s",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_S)
MLE_KEY_DEF("t",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_T)
MLE_KEY_DEF("u",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_U)
MLE_KEY_DEF("v",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_V)
MLE_KEY_DEF("w",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_W)
MLE_KEY_DEF("x",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_X)
MLE_KEY_DEF("y",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_Y)
MLE_KEY_DEF("z",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_Z)
MLE_KEY_DEF("4",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_4)
MLE_KEY_DEF("\\",        TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_4)
MLE_KEY_DEF("5",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_5)
MLE_KEY_DEF("6",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_6)
MLE_KEY_DEF("7",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_7)
MLE_KEY_DEF("/",         TB_MOD_CTRL, 0,           0,  TB_KEY_CTRL_SLASH)
MLE_KEY_DEF("backspace", 0,           TB_MOD_CTRL, 0,  TB_KEY_CTRL_8)
MLE_KEY_DEF("escape",    0,           0,           0,  TB_KEY_ESC)
MLE_KEY_DEF("space",     0,           0,           32, 0)
MLE_KEY_DEF("comma",     0,           0,           44, 0)
MLE_KEY_DEF("up",        0,           0,           0,  TB_KEY_ARROW_UP)
MLE_KEY_DEF("down",      0,           0,           0,  TB_KEY_ARROW_DOWN)
MLE_KEY_DEF("left",      0,           0,           0,  TB_KEY_ARROW_LEFT)
MLE_KEY_DEF("right",     0,           0,           0,  TB_KEY_ARROW_RIGHT)
MLE_KEY_DEF("insert",    0,           0,           0,  TB_KEY_INSERT)
MLE_KEY_DEF("delete",    0,           0,           0,  TB_KEY_DELETE)
MLE_KEY_DEF("home",      0,           0,           0,  TB_KEY_HOME)
MLE_KEY_DEF("end",       0,           0,           0,  TB_KEY_END)
MLE_KEY_DEF("pgup",      0,           0,           0,  TB_KEY_PGUP)
MLE_KEY_DEF("pgdn",      0,           0,           0,  TB_KEY_PGDN)
MLE_KEY_DEF("backtab",   0,           0,           0,  TB_KEY_BACK_TAB)
MLE_KEY_DEF("f1",        0,           0,           0,  TB_KEY_F1)
MLE_KEY_DEF("f2",        0,           0,           0,  TB_KEY_F2)
MLE_KEY_DEF("f3",        0,           0,           0,  TB_KEY_F3)
MLE_KEY_DEF("f4",        0,           0,           0,  TB_KEY_F4)
MLE_KEY_DEF("f5",        0,           0,           0,  TB_KEY_F5)
MLE_KEY_DEF("f6",        0,           0,           0,  TB_KEY_F6)
MLE_KEY_DEF("f7",        0,           0,           0,  TB_KEY_F7)
MLE_KEY_DEF("f8",        0,           0,           0,  TB_KEY_F8)
MLE_KEY_DEF("f9",        0,           0,           0,  TB_KEY_F9)
MLE_KEY_DEF("f10",       0,           0,           0,  TB_KEY_F10)
MLE_KEY_DEF("f11",       0,           0,           0,  TB_KEY_F11)
MLE_KEY_DEF("f12",       0,           0,           0,  TB_KEY_F12)
